# -*- coding: utf-8 -*-

from .agent.api import DEFAULT_USERNAME
from .agent.api import DEFAULT_PASSWORD
from .agent.api import DEFAULT_HOST
from .agent.api import DEFAULT_PORT
from .agent.api import SOAPRequest
from .agent.api import SOAPResponse
from .sdk.api import run_soap_command
from .sdk.api import canned



# -*- coding: utf-8 -*-

"""
对具体的 GM 命令业务逻辑高度封装后的接口. 这些接口可以分为两类:

1. 我们只关心运行成功与否, 不关心返回值.
2. 我们需要这个命令返回的数据.

所有的接口都有这三个参数:

- bsm: ``boto_session_manager.BotoSesManager`` 对象, 定义了 AWS 权限.
- server_id: 服务器的逻辑 ID, 命名规则为 ``${env_name}-${server_name}``. 例如 ``sbx-blue``.
- raises: 当命令执行失败时是否抛出异常, 默认为 ``True``.

Reference:

- https://www.azerothcore.org/wiki/gm-commands
"""

import typing as T
import re
from boto_session_manager import BotoSesManager

from ...exc import SOAPResponseParseError, SOAPCommandFailedError
from ..core import run_soap_command


def extract_online_players(message: str) -> T.Tuple[int, int]:
    res = re.findall(r"Connected players: (\d+)", message)
    if len(res) == 1:
        connected_players = int(res[0])
    else:  # pragma: no cover
        raise SOAPResponseParseError(message)

    res = re.findall(r"Characters in world: (\d+)", message)
    if len(res) == 1:
        characters_in_world = int(res[0])
    else:  # pragma: no cover
        raise SOAPResponseParseError(message)

    return connected_players, characters_in_world


def get_online_players(
    bsm: BotoSesManager,
    server_id: str,
    raises: bool = True,
) -> T.Dict[str, int]:
    """
    :return: a dict with two keys: ``connected_players`` and ``characters_in_world``.
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=".server info",
        raises=raises,
    )[0]

    connected_players, characters_in_world = extract_online_players(response.message)

    return {
        "connected_players": connected_players,
        "characters_in_world": characters_in_world,
    }


def is_server_online(
    bsm: BotoSesManager,
    server_id: str,
    raises: bool = True,
) -> bool:
    """
    :return: a boolean value to indicate whether the server is online
    """
    result = get_online_players(bsm, server_id, raises=raises)
    return True


def create_account(
    bsm: BotoSesManager,
    server_id: str,
    username: str,
    password: str,
    raises: bool = True,
) -> bool:
    """
    :return: a boolean value to indicate whether the account is created successfully
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=f".account create {username} {password}",
        raises=raises,
    )[0]
    response.print()
    return response.succeeded


def set_gm_level(
    bsm: BotoSesManager,
    server_id: str,
    username: str,
    level: int,
    realm_id: int,
    raises: bool = True,
) -> bool:
    """
    :param username:
    :param level:
    :param realm_id:

    :return: a boolean value to indicate whether the account is created successfully
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=f".account set gmlevel {username} {level} {realm_id}",
        raises=raises,
    )[0]
    response.print()
    return response.succeeded


def set_password(
    bsm: BotoSesManager,
    server_id: str,
    username: str,
    password: str,
    raises: bool = True,
) -> bool:
    """
    :param username:
    :param level:
    :param realm_id:

    :return: a boolean value to indicate whether the account is created successfully
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=f".account set password {username} {password} {password}",
        raises=raises,
    )[0]
    response.print()
    return response.succeeded


def delete_account(
    bsm: BotoSesManager,
    server_id: str,
    username: str,
    raises: bool = True,
) -> bool:
    """
    :param username:

    :return: a boolean value to indicate whether the account is deleted successfully
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=f".account delete {username}",
        raises=raises,
    )[0]
    response.print()
    return response.succeeded


def gm_list(
    bsm: BotoSesManager,
    server_id: str,
    raises: bool = True,
) -> T.List[T.Tuple[str, int]]:
    """
    :return: a boolean value to indicate whether the account is deleted successfully
    """
    response = run_soap_command(
        bsm=bsm,
        server_id=server_id,
        request_like=f".gm list",
        raises=raises,
    )[0]
    if response.succeeded:
        results = list()
        lines = response.message.splitlines()
        for line in lines:
            if line.startswith("|"):
                words = [word.strip() for word in line.split("|") if word.strip()]
                results.append((words[0], int(words[1])))
        return results
    else:
        raise SOAPCommandFailedError



# -*- coding: utf-8 -*-

"""
Agent implementation.
"""

import typing as T
import json
import dataclasses
from pathlib import Path
from datetime import datetime
import xml.etree.ElementTree as ET

import boto3
import requests

from ..paths import dir_python_lib
from ..exc import SOAPResponseParseError
from ..utils import get_object, put_object


# ------------------------------------------------------------------------------
# Soap Request and Response
# ------------------------------------------------------------------------------
path_xml = dir_python_lib.joinpath("agent", "execute-command.xml")

# default soap request headers
_SOAP_REQUEST_HEADERS = {"Content-Type": "application/xml"}
_SOAP_REQUEST_XML_TEMPLATE = path_xml.read_text(encoding="utf-8")
DEFAULT_USERNAME = "admin"
DEFAULT_PASSWORD = "admin"
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 7878


@dataclasses.dataclass
class Base:
    """
    Base class for :class:`SOAPRequest` and :class:`SOAPResponse`.
    """
    @classmethod
    def from_dict(cls, dct: dict):
        """
        Construct an object from a dict.
        """
        return cls(**dct)

    def to_dict(self) -> dict:
        """
        Convert the object to a dict.
        """
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}

    @classmethod
    def from_json(cls, json_str: str):
        """
        Construct an object from a JSON string.
        """
        return cls.from_dict(json.loads(json_str))

    def to_json(self) -> str:  # pragma: no cover
        """
        Convert the object to a JSON string.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def batch_load_from_s3(
        cls,
        s3_client,
        s3uri: str,
    ):
        """
        从 S3 中加载多个对象.
        """
        json_str = get_object(s3_client, s3uri=s3uri)
        return [cls.from_dict(dct) for dct in json.loads(json_str)]

    @classmethod
    def batch_dump_to_s3(
        cls,
        s3_client,
        instances: T.Iterable,
        s3uri: str,
    ):
        """
        将多个对象以 JSON 格式保存到 S3 中.
        """
        data = [instance.to_dict() for instance in instances]
        put_object(s3_client=s3_client, s3uri=s3uri, body=json.dumps(data))


@dataclasses.dataclass
class SOAPRequest(Base):
    """
    :class:`~acore_soap_app.agent.impl.SOAPRequest` is a dataclass to represent
    the SOAP XML request.

    Usage example

    .. code-block:: python

        # this code only works in EC2 environment
        >>> request = SOAPRequest(command=".server info")
        >>> response = request.send()
        >>> response.to_json()
        {
            "body": "<?xml version=\"1.0\" encoding=\"UTF-8\"?><SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:SOAP-ENC=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:xsi=\"http://www.w3.org/1999/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/1999/XMLSchema\" xmlns:ns1=\"urn:AC\"><SOAP-ENV:Body><ns1:executeCommandResponse><result>AzerothCore rev. 85311fa55983 2023-03-25 22:36:05 +0000 (master branch) (Unix, RelWithDebInfo, Static)&#xD;Connected players: 0. Characters in world: 0.&#xD;Connection peak: 0.&#xD;Server uptime: 54 minute(s) 3 second(s)&#xD;Update time diff: 10ms, average: 10ms.&#xD;</result></ns1:executeCommandResponse></SOAP-ENV:Body></SOAP-ENV:Envelope>",
            "message": "AzerothCore rev. 85311fa55983 2023-03-25 22:36:05 +0000 (master branch) (Unix, RelWithDebInfo, Static)Connected players: 0. Characters in world: 0.Connection peak: 0.Server uptime: 54 minute(s) 3 second(s)Update time diff: 10ms, average: 10ms.",
            "succeeded": true
        }

    :param command: the command to execute.
    :param username: the in game GM account username, default "admin".
    :param password: the in game GM account password, default "admin".
    :param host: wow world server host, default "localhost".
    :param port: wow world server SOAP port, default 7878.

    More methods from base class:

    - :meth:`~Base.from_dict`
    - :meth:`~Base.to_dict`
    - :meth:`~Base.from_json`
    - :meth:`~Base.to_json`
    - :meth:`~Base.batch_load_from_s3`
    - :meth:`~Base.batch_dump_to_s3`
    """

    command: str = dataclasses.field()
    username: T.Optional[str] = dataclasses.field(default=None)
    password: T.Optional[str] = dataclasses.field(default=None)
    host: T.Optional[str] = dataclasses.field(default=None)
    port: T.Optional[int] = dataclasses.field(default=None)

    def set_default(
        self,
        username: T.Optional[str],
        password: T.Optional[str],
    ):
        """
        Set default values for username and password if they are not set.
        """
        if self.username is None:
            self.username = username
        if self.password is None:
            self.password = password

    @property
    def endpoint(self) -> str:
        """
        Construct the Soap service endpoint URL.
        """
        username = self.username or DEFAULT_USERNAME
        password = self.password or DEFAULT_PASSWORD
        host = self.host or DEFAULT_HOST
        port = self.port or DEFAULT_PORT
        return f"http://{username}:{password}@{host}:{port}/"

    def send(self) -> "SOAPResponse":  # pragma: no cover
        """
        Run soap command via HTTP request. This function "has to" be run on the
        game server and talk to the localhost. You should NEVER open SOAP port
        to public!
        """
        http_response = requests.post(
            self.endpoint,
            headers=_SOAP_REQUEST_HEADERS,
            data=_SOAP_REQUEST_XML_TEMPLATE.format(command=self.command),
        )
        return SOAPResponse.parse(http_response.text)

    @classmethod
    def batch_load(
        cls,
        request_like: T.Union[
            str,
            T.List[str],
            "SOAPRequest",
            T.List["SOAPRequest"],
        ],
        username: T.Optional[str] = None,
        password: T.Optional[str] = None,
        s3_client=None,
    ) -> T.List["SOAPRequest"]:
        """
        从各种形式的输入中加载 :class:`SOAPRequest`. 该方法总是返回一个列表.

        输入参数 ``request_like`` 代表着要运行的 GM 命令, 它可能是以下几种形式中的一种:

        - 如果是一个字符串:
            - 如果是以 s3:// 开头, 那么就去 S3 读数据, 此时需要给定 ``s3_client`` 参数.
                通常用于 payload 比较大的情况:
                - 如果读到的数据是单个字典, 那么就视为一个 SOAPRequest.
                - 如果读到的数据是一个列表, 那么就视为多个 SOAPRequest.
            - 如果不是以 s3:// 开头, 那么就视为一个 GM command.
        - 如果是一个字符串列表, 那么就视为多个 GM 命令.
        - 它还可以是单个 SOAPRequest 或 SOAPRequest 列表.
        - 这个参数最终都会被转换成 SOAPRequest 的列表.

        :param request_like: 上面已经说过了.
        :param username: 默认的用户名, 只有当 request.username 为 None 的时候才会用到.
        :param password: 默认的密码, 只有当 request.password 为 None 的时候才会用到.
        :param s3_client: boto3.client("s3")
        """
        if isinstance(request_like, str):
            if request_like.startswith("s3://"):  # pragma: no cover
                data = json.loads(get_object(s3_client=s3_client, s3uri=request_like))
                if isinstance(data, dict):  # pragma: no cover
                    requests = [SOAPRequest.from_dict(data)]
                elif isinstance(data, list):
                    requests = [SOAPRequest.from_dict(dct) for dct in data]
                else:  # pragma: no cover
                    raise TypeError(
                        f"data in S3 must be a dict or "
                        f"a list of dict, not {type(data)}"
                    )
            else:
                requests = [SOAPRequest(command=request_like)]
        elif isinstance(request_like, SOAPRequest):
            requests = [request_like]
        elif isinstance(request_like, list):
            if isinstance(request_like[0], str):
                requests = [SOAPRequest(command=item) for item in request_like]
            elif isinstance(request_like[0], SOAPRequest):
                requests = request_like
            else:
                raise TypeError(
                    f"item ``request_like`` list must be "
                    f"str or SOAPRequest, not {type(request_like[0])}"
                )
        else:  # pragma: no cover
            raise TypeError(
                f"request must be str or SOAPRequest, " f"not {type(request_like)}"
            )
        for request in requests:
            request.set_default(username=username, password=password)
        return requests


@dataclasses.dataclass
class SOAPResponse(Base):
    """
    :class:`~acore_soap_app.agent.impl.SOAPResponse` is a dataclass to represent
    the SOAP XML response.

    Usage:

    .. code-block:: python

        >>> res = SOAPResponse.parse(
        ...  '''
        ...      <?xml version="1.0" encoding="UTF-8"?><SOAP-ENV:Envelope
        ...      ...<result>Account created: test&#xD;</result>...</SOAP-ENV:Envelope>
        ...  '''
        ... )
        >>> res.message
        Account created: test
        >>> res.succeeded
        True

    :param body: the raw SOAP XML response
    :param message: if succeeded, it is the ``<result>...</result>`` part.
        if failed, it is the ``<faultstring>...</faultstring>`` part
    :param succeeded: a boolean flag to indicate whether the command is succeeded

    More methods from base class:

    - :meth:`~Base.from_dict`
    - :meth:`~Base.to_dict`
    - :meth:`~Base.from_json`
    - :meth:`~Base.to_json`
    - :meth:`~Base.batch_load_from_s3`
    - :meth:`~Base.batch_dump_to_s3`
    """

    body: str = dataclasses.field()
    message: str = dataclasses.field()
    succeeded: bool = dataclasses.field()

    @classmethod
    def parse(cls, body: str) -> "SOAPResponse":
        """
        Parse the SOAP XML response.
        """
        root = ET.fromstring(body)
        results = list(root.iter("result"))
        if len(results):
            return cls(
                body=body.strip(),
                message=results[0].text.strip(),
                succeeded=True,
            )
        faultstrings = list(root.iter("faultstring"))
        if len(faultstrings):
            return cls(
                body=body.strip(),
                message=faultstrings[0].text.strip(),
                succeeded=False,
            )

        raise SOAPResponseParseError(f"Cannot parse the response: {body!r}")

    def print(self):  # pragma: no cover
        """
        Print the dataclass, ignore the raw response body.
        """
        print({"succeeded": self.succeeded, "message": self.message})


# ------------------------------------------------------------------------------
# Create boto3 session on EC2
# ------------------------------------------------------------------------------
path_aws_region_cache = Path.home().joinpath(".aws_region_cache.json")


def _dump_aws_region_cache(region: str):  # pragma: no cover
    path_aws_region_cache.write_text(
        json.dumps({"region": region, "timestamp": int(datetime.utcnow().timestamp())})
    )


def _load_aws_region_cache() -> T.Optional[str]:  # pragma: no cover
    if not path_aws_region_cache.exists():
        return None
    data = json.loads(path_aws_region_cache.read_text())
    if (datetime.utcnow().timestamp() - data["timestamp"]) > 86400:  # cache expired
        return None
    else:
        return data["region"]


def get_ec2_metadata(name: str) -> str:  # pragma: no cover
    """
    Get the EC2 instance id from the AWS EC2 metadata API.

    Reference:

    - https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
    """
    url = f"http://169.254.169.254/latest/meta-data/{name}"
    return requests.get(url).text.strip()


def get_ec2_region() -> str:  # pragma: no cover
    region = _load_aws_region_cache()
    if region is None:
        region = get_ec2_metadata("placement/region")
        _dump_aws_region_cache(region)
    return region


def get_boto_ses() -> boto3.session.Session:  # pragma: no cover
    """
    On EC2, we use the IAM role to get the AWS credentials. You only need to
    specify the region name, however, it is automatically discovered by the
    EC2 metadata API.
    """
    return boto3.session.Session(region_name=get_ec2_region())



# -*- coding: utf-8 -*-

"""
SOAP agent is a command line tool on the EC2 server that can run GM command
via SOAP protocol and return the SOAP XML response in JSON format.

With a SOAP agent installed, developer can use AWS SSM run command feature to
run this CLI command remotely from anywhere (if have the right permission).

[CN]

SOAP Agent 是 一个在 EC2 服务器上的命令行工具 (你需要 pip install 安装), 它可以通过
SOAP 协议来运行 GM 命令, 并获得 JSON 格式的响应. SOAP Agent 是游戏内 GM 运营自动化的核心,
它提供了一个接口, 使得开发者可以在任何有权限的地方, 通过 AWS SSM run command 的功能, 执行
单个或批量 GM 命令. 如果输入和输出都很大, 还可以使用 AWS S3 作为数据储存的媒介.
"""



# -*- coding: utf-8 -*-

from acore_soap_app import api


def test():
    _ = api


if __name__ == "__main__":
    from acore_soap_app.tests import run_cov_test

    run_cov_test(__file__, "acore_soap_app.api", preview=False)



# -*- coding: utf-8 -*-

"""
agent module public API.

Example:

.. code-block:: python

    >>> from acore_soap_app.agent.api import DEFAULT_USERNAME
    >>> from acore_soap_app.agent.api import DEFAULT_PASSWORD
    >>> from acore_soap_app.agent.api import DEFAULT_HOST
    >>> from acore_soap_app.agent.api import DEFAULT_PORT
    >>> from acore_soap_app.agent.api import SOAPRequest
    >>> from acore_soap_app.agent.api import SOAPResponse
    >>> from acore_soap_app.agent.api import get_boto_ses
"""

from .impl import DEFAULT_USERNAME
from .impl import DEFAULT_PASSWORD
from .impl import DEFAULT_HOST
from .impl import DEFAULT_PORT
from .impl import SOAPRequest
from .impl import SOAPResponse
from .impl import get_boto_ses



# -*- coding: utf-8 -*-


def get_object(s3_client, s3uri: str) -> str:
    """
    从 S3 中读取一个对象, 并返回其文本内容.
    """
    parts = s3uri.split("/", 3)
    bucket, key = parts[2], parts[3]
    response = s3_client.get_object(Bucket=bucket, Key=key)
    return response["Body"].read().decode("utf-8")


def put_object(s3_client, s3uri: str, body: str):
    """
    将一个 JSON 对象保存到 S3 中.
    """
    parts = s3uri.split("/", 3)
    bucket, key = parts[2], parts[3]
    return s3_client.put_object(
        Bucket=bucket,
        Key=key,
        Body=body,
        ContentType="application/json",
    )



# -*- coding: utf-8 -*-

"""
Command line low lever implementations.
"""

import typing as T
import json
from datetime import datetime, timezone

from acore_constants.api import TagKey

from ..agent.api import SOAPRequest, SOAPResponse, get_boto_ses
from ..sdk.api import canned
from ..exc import SOAPCommandFailedError


def ensure_ec2_environment():
    if not __file__.startswith("/home/ubuntu"):
        raise EnvironmentError("This is not a EC2 environment")


def ensure_response_succeeded(
    request: SOAPRequest,
    response: SOAPResponse,
    raises: bool,
) -> SOAPResponse:
    if response.succeeded:
        return response
    else:
        if raises:
            raise SOAPCommandFailedError(
                f"request failed: {request.command!r}, "
                f"response: {response.message!r}"
            )
        else:
            return response


def gm(
    request_like: T.Union[
        str,
        T.List[str],
        "SOAPRequest",
        T.List["SOAPRequest"],
    ],
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
    raises: bool = True,
    s3uri_output: T.Optional[str] = None,
):
    """
    运行一个或多个 GM 命令. 例如 ``.server info``.

    :param request_like: 请参考 :class:`~acore_soap_app.agent.impl.SOAPRequest.batch_load`
    :param username: 默认的用户名, 只有当 request.username 为 None 的时候才会用到.
    :param password: 默认的密码, 只有当 request.password 为 None 的时候才会用到.
    :param raises: 默认为 True. 如果为 True, 则在遇到错误时抛出异常. 反之则将
        failed SOAP Response 原封不动地返回.
    :param s3uri_output: 可选参数, 如果为 None, 则将
        :class:`~acore_soap_app.agent.impl.SOAPResponse` 对象转换为 JSON 并打印.
        如果给定, 则将 JSON 保存到 S3 中. 常用于返回结果特别大的情况.
    """
    ensure_ec2_environment()

    # handle input
    boto_ses = get_boto_ses()
    s3_client = boto_ses.client("s3")
    requests = SOAPRequest.batch_load(
        request_like=request_like,
        username=username,
        password=password,
        s3_client=s3_client,
    )

    # run
    responses = [
        ensure_response_succeeded(request, request.send(), raises=raises)
        for request in requests
    ]

    # handle output
    if s3uri_output is None:
        print("\n".join([response.to_json() for response in responses]))
    else:
        SOAPResponse.batch_dump_to_s3(
            s3_client=s3_client,
            instances=responses,
            s3uri=s3uri_output,
        )


def _count_online_players(
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
) -> dict:
    """
    底层函数, 通过运行 GM 命令来获得服务器的在线玩家数, 同时获得服务器的在线状态.
    返回一个包含以上信息的字典.

    Example output:

    .. code-block:: python

        {
            "connected_players": 1000,
            "characters_in_world": 700,
            "server_is_online": True,
        }
    """
    requests = SOAPRequest.batch_load(
        request_like=".server info",
        username=username,
        password=password,
    )
    request = requests[0]
    try:
        response = request.send()
        connected_players, characters_in_world = canned.extract_online_players(
            response.message
        )
        return {
            "connected_players": connected_players,
            "characters_in_world": characters_in_world,
            "server_is_online": True,
        }
    except Exception as e:
        return {
            "connected_players": None,
            "characters_in_world": None,
            "server_is_online": False,
        }


def count_online_players(
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
):
    """
    打印在线玩家数量, 以及服务器的在线状态, 以 JSON 的形式输出到 stdout.
    """
    data = _count_online_players(username=username, password=password)
    print(json.dumps(data, indent=4))


def measure_server_status(
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
):
    """
    测量服务器的在线状态, 并将结果保存到 EC2 实例的 Tags 中.
    """
    from boto_session_manager import BotoSesManager
    from simple_aws_ec2.api import Ec2Instance

    data = _count_online_players(username=username, password=password)
    now = datetime.utcnow().replace(tzinfo=timezone.utc)

    if data["server_is_online"]:
        tags = {
            TagKey.WOW_STATUS: "{} players".format(data["connected_players"]),
            TagKey.WOW_STATUS_MEASURE_TIME: now.isoformat(),
        }
    else:
        tags = {
            TagKey.WOW_STATUS: "stopped",
            TagKey.WOW_STATUS_MEASURE_TIME: now.isoformat(),
        }

    aws_region = Ec2Instance.get_placement_region()
    bsm = BotoSesManager(region_name=aws_region)
    instance_id = Ec2Instance.get_instance_id()
    bsm.ec2_client.create_tags(
        Resources=[instance_id],
        Tags=[dict(Key=k, Value=v) for k, v in tags.items()],
    )



# -*- coding: utf-8 -*-

from .impl import extract_online_players
from .impl import get_online_players
from .impl import is_server_online
from .impl import create_account
from .impl import set_gm_level
from .impl import set_password
from .impl import delete_account
from .impl import gm_list



# -*- coding: utf-8 -*-

"""
SOAP Agent command line user interface.
"""

import typing as T
import fire

from .impl import (
    gm,
    count_online_players,
    measure_server_status,
)


class Canned:
    """
    A collection of canned SOAP Agent commands.
    """
    def count_online_players(
        self,
        user: T.Optional[str] = None,
        pwd: T.Optional[str] = None,
    ):
        """
        Get the online players and characters in world. Also, you can use this
         command to check whether server is online.

        Example::

            acsoap canned count-online-players --help

            acsoap canned count-online-players
        """
        count_online_players(
            username=user,
            password=pwd,
        )

    def measure_server_status(
        self,
        user: T.Optional[str] = None,
        pwd: T.Optional[str] = None,
    ):
        """
        Measure server status and save result to EC2 tags.

        Example::

            acsoap canned measure-server-status --help

            acsoap canned measure-server-status
        """
        measure_server_status(
            username=user,
            password=pwd,
        )



class Command:
    """
    Acore Soap Agent command line interface. All these commands can only be
    used on EC2.

    Example:

    - acsoap
    """

    def __init__(self):
        self.canned = Canned()

    def gm(
        self,
        cmd: str,
        user: T.Optional[str] = None,
        pwd: T.Optional[str] = None,
        raises: bool = True,
        s3uri: T.Optional[str] = None,
    ):
        """
        Run single GM command.

        :param cmd: the GM command to run
        :param user: in game GM account username, if not given, then use "admin"
        :param pwd: in game GM account password, if not given, then use "admin"
        :param raises: raise error if any of the GM command failed.
        :param s3uri: if None, then return the response as JSON, otherwise, save
            the response to S3.

        Example::

            acsoap gm --help

            acsoap gm ".server info"

            acsoap gm ".server info" --user myuser --pwd mypwd

            acsoap gm ".server info" --s3uri s3://bucket/output.json
        """
        gm(
            request_like=cmd,
            username=user,
            password=pwd,
            raises=raises,
            s3uri_output=s3uri,
        )


def run():
    fire.Fire(Command)



# -*- coding: utf-8 -*-

"""
这是 SDK 的核心模块, 实现了各种底层方法.
"""

import typing as T
import time

from boto_session_manager import BotoSesManager
import aws_ssm_run_command.api as aws_ssm_run_command
from acore_server_metadata.api import Server

from ..agent.api import SOAPRequest, SOAPResponse
from ..exc import EC2IsNotRunningError, RunCommandError


def build_cli_arg_for_gm(
    command: str,
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
    raises: bool = True,
    s3uri_output: T.Optional[str] = None,
    path_cli: str = "/home/ubuntu/git_repos/acore_soap_app-project/.venv/bin/acsoap",
) -> str:
    """
    构造最终的命令行参数.
    """
    args = [
        path_cli,
        "gm",
        f'"{command}"',
    ]
    if username is not None:
        args.append(f"-u={username}")
    if password is not None:
        args.append(f"-p={password}")
    if raises is not None:
        if raises:
            args.append(f"-r=True")
        else:
            args.append(f"-r=False")
    if s3uri_output is not None:
        args.append(f"-s={s3uri_output}")
    return " ".join(args)


def run_soap_command(
    bsm: BotoSesManager,
    server_id: str,
    request_like: T.Union[
        str,
        T.List[str],
        SOAPRequest,
        T.List[SOAPRequest],
    ],
    username: T.Optional[str] = None,
    password: T.Optional[str] = None,
    raises: bool = True,
    s3uri_input: T.Optional[str] = None,
    s3uri_output: T.Optional[str] = None,
    path_cli: str = "/home/ubuntu/git_repos/acore_soap_app-project/.venv/bin/acsoap",
    sync: bool = True,
    delays: int = 1,
    timeout: int = 10,
    verbose: bool = True,
) -> T.Union[T.List[SOAPResponse], str]:
    """
    从任何地方, 通过 SSM Run Command, 远程执行 SOAP 命令.

    Usage Example:

    .. code-block:: python

        >>> response = run_soap_command(bsm, "sbx-blue", ".server info")
        >>> response = run_soap_command(bsm, "sbx-blue", [".account create test1 test1", ".account create test2 test2"])

    :param bsm: ``boto_session_manager.BotoSesManager`` 对象, 定义了 AWS 权限.
    :param server_id: AzerothCore 服务器的逻辑 ID, 命名规律为 "${env_name}-${server_name}",
        例如 "sbx-blue"
    :param request_like: 请参考
        :class:`~acore_soap_app.agent.impl.SOAPRequest.batch_load`
    :param username: 默认的用户名, 只有当 request.username 为 None 的时候才会用到.
    :param password: 默认的密码, 只有当 request.password 为 None 的时候才会用到.
    :param raises: 默认为 True. 如果为 True, 则在遇到错误时抛出异常. 反之则将
        failed SOAP Response 原封不动地返回.
    :param s3uri_input: 如果指定, 则将输入写入 S3. 常用于 Payload 比较大的情况.
        如果你一次性发送的 request 大于 20 条, 则必须使用这个参数.
    :param s3uri_output: 如果不指定, 则默认将输出作为 JSON 打印. 如果指定了 s3uri,
        则将输出写入到 S3.
    :param path_cli: EC2 上 acsoap 命令行工具的绝对路径.
    :param sync: 同步和异步模式, 默认为同步模式
        - 如果以同步模式运行, 则会等待 SSM Run Command 完成
        - 如果以异步模式运行, 则会立刻返回一个 SSM Run Command 的 command id.
    :param delays: 同步模式下的等待间隔
    :param timeout: 同步模式下的超时限制
    :param verbose: 同步模式下是否显示进度条
    """
    # load requests
    requests = SOAPRequest.batch_load(
        request_like=request_like,
        username=username,
        password=password,
        s3_client=bsm.s3_client,
    )

    # get ec2 instance id
    server = Server(id=server_id)
    server.refresh(ec2_client=bsm.ec2_client, rds_client=bsm.rds_client)
    if server.is_running() is False:
        raise EC2IsNotRunningError(f"EC2 {server_id!r} is not running")
    instance_id = server.ec2_inst.id

    # identify the run strategy
    if len(requests) >= 20:
        if s3uri_input is None:
            raise ValueError(
                "'s3uri_input' must be specified when the number of requests is greater than 20"
            )
        is_s3_input = True
    else:
        is_s3_input = s3uri_input is not None

    if is_s3_input:
        SOAPRequest.batch_dump_to_s3(
            s3_client=bsm.s3_client,
            instances=requests,
            s3uri=s3uri_input,
        )
        commands = [
            build_cli_arg_for_gm(
                command=s3uri_input,
                username=username,
                password=password,
                raises=raises,
                s3uri_output=s3uri_output,
                path_cli=path_cli,
            )
        ]
    else:
        commands = [
            build_cli_arg_for_gm(
                command=request.command,
                username=request.username,
                password=request.password,
                raises=raises,
                s3uri_output=s3uri_output,
                path_cli=path_cli,
            )
            for request in requests
        ]

    # run command
    command_id = aws_ssm_run_command.better_boto.send_command(
        ssm_client=bsm.ssm_client,
        instance_id=instance_id,
        commands=commands,
    )
    if sync is False:  # async mode, return immediately
        return command_id

    # sync mode, wait until command succeeded
    time.sleep(1)

    # get command response
    command_invocation = aws_ssm_run_command.better_boto.wait_until_command_succeeded(
        ssm_client=bsm.ssm_client,
        command_id=command_id,
        instance_id=instance_id,
        raises=False,
        delays=delays,
        timeout=timeout,
        verbose=verbose,
    )
    if command_invocation.ResponseCode != 0:
        raise RunCommandError.from_command_invocation(command_invocation)

    # parse response
    if s3uri_output is None:
        output = command_invocation.StandardOutputContent
        lines = output.splitlines()
        responses = [SOAPResponse.from_json(json_str) for json_str in lines]
    else:
        responses = SOAPResponse.batch_load_from_s3(
            s3_client=bsm.s3_client, s3uri=s3uri_output
        )
    return responses



# -*- coding: utf-8 -*-

import typing as T


if T.TYPE_CHECKING:
    import aws_ssm_run_command.api as aws_ssm_run_command


class EC2IsNotRunningError(SystemError):
    """
    raises when try to run a remote command on a non-running EC2 instance.
    """

    pass


class RunCommandError(SystemError):
    """
    raises when ssm_client.send_command() fails.
    """

    @classmethod
    def from_command_invocation(
        cls,
        command_invocation: "aws_ssm_run_command.better_boto.CommandInvocation",
    ):
        msg = (
            f"return code: {command_invocation.ResponseCode}, "
            f"output: {command_invocation.StandardOutputContent!r}, "
            f"error: {command_invocation.StandardErrorContent!r}, "
        )
        return cls(msg)


class SOAPResponseParseError(ValueError):
    """
    raises when failed to parse the soap response.
    """


class SOAPCommandFailedError(ValueError):
    """
    raises when SOAP command failed
    """



# -*- coding: utf-8 -*-

"""
sdk module public API.

Example:

.. code-block:: python

    >>> from acore_soap_app.sdk.api import canned
    >>> from acore_soap_app.sdk.api import run_soap_command
"""

from .core import run_soap_command
from .canned import api as canned
